from ignis import CACHE_DIR
from gi.repository import GLib  # type: ignore
from ignis.options_manager import OptionsManager, OptionsGroup, TrackedList
from ignis import is_sphinx_build


def get_recorder_default_file_location() -> str | None:
    if not is_sphinx_build:
        return GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_VIDEOS)
    else:
        return "XDG Videos directory"


class Options(OptionsManager):
    """
    Options for Ignis.

    .. warning::

        Use already initialized instance of this class:

        .. code-block:: python

            from ignis.options import options

            print(options.notifications.dnd)

    Below are classes with options, their names begin with a capital letter.
    However, if you want to get the current value of an option or set a value,
    use an initialized instance that starts with a lowercase letter.

    For example:
        * ``Notifications`` -> ``notifications``
        * ``Recorder`` -> ``recorder``
        * and etc.

    You can use classes (not instances of them) to obtain default values of options.

    .. hint::
        If the option is of type :class:`~ignis.options_manager.TrackedList`, it means that it is regular Python list.
        But you can call ``.append()``, ``.remove()``, ``.insert()``, etc., and the changes will be applied!

    Example usage:

    .. code-block::

        from ignis.options import options

        # Get an option value
        print(options.notifications.dnd)

        # Set a new value for an option
        options.notifications.dnd = True

        # Connect to an option change event
        options.notifications.connect_option("dnd", lambda: print("option dnd changed! new value:", options.notifications.dnd))

        # You can also bind to an option!
        options.notifications.bind("dnd")

        # Obtain the default value of an option
        print(options.Notifications.popup_timeout)
    """

    def __init__(self):
        if is_sphinx_build:
            return

        try:
            super().__init__(file=f"{CACHE_DIR}/ignis_options.json")
        except FileNotFoundError:
            pass

    class Notifications(OptionsGroup):
        """
        Options for the :class:`~ignis.services.notifications.NotificationService`.
        """

        #: Do Not Disturb mode.
        #:
        #: If set to ``True``, the ``new_popup`` signal will not be emitted,
        #: and all new :class:`~ignis.services.notifications.Notification` instances will have ``popup`` set to ``False``.
        dnd: bool = False

        #: The timeout before a popup is automatically dismissed, in milliseconds.
        popup_timeout: int = 5000

        #: The maximum number of popups.
        #:
        #: If the length of the ``popups`` list exceeds ``max_popups_count``, the oldest popup will be dismissed.
        max_popups_count: int = 3

    class Recorder(OptionsGroup):
        """
        Options for the :class:`~ignis.services.recorder.RecorderService`.
        """

        #: The bitrate of the recording.
        bitrate: int = 8000

        #: The default location for saving recordings. Defaults to XDG Video directory.
        default_file_location: str | None = get_recorder_default_file_location()

        #: The default filename for recordings. Supports time formating.
        default_filename: str = "%Y-%m-%d_%H-%M-%S.mp4"

    class Applications(OptionsGroup):
        """
        Options for the :class:`~ignis.services.applications.ApplicationsService`.
        """

        #: A list of the pinned applications desktop files, e.g. ``"firefox.desktop"``, ``"code.desktop"``.
        pinned_apps: TrackedList[str] = TrackedList()

    class Wallpaper(OptionsGroup):
        """
        Options for the :class:`~ignis.services.wallpaper.WallpaperService`.
        """

        #: The path to the wallpaper image.
        wallpaper_path: str | None = None

    notifications = Notifications()
    recorder = Recorder()
    applications = Applications()
    wallpaper = Wallpaper()


options = Options()
